package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.ClickhouseDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.PlanStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfHourODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.PlanStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.entity.PlanStatOfHourDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.PlanUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.windows.GlobalWindow;
import org.apache.flink.util.Collector;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;

import java.time.LocalDate;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;

@Slf4j
public class PlanHourDWDProcess extends KeyedProcessFunction<Long, AdDataOfHourODS, PlanStatOfHourDWD> {

    private SqlSessionFactory sqlSessionFactory;
    // 上次查询的天数据
    private ValueState<String> lastQueryDayState;
    // 聚合的天的数据
    private MapState<String, PlanStatOfDayDWD> historyReduceState;
    // 昨天集合的最后一次数据
    private MapState<String, PlanStatOfHourDWD> lastReduceState;

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties ckProps = new Properties();
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_URL, params.get(ApplicationProperties.CK_URL));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_USER, params.get(ApplicationProperties.CK_USERNAME));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_PASSWORD, params.get(ApplicationProperties.CK_PASSWORD));

        DataSourceFactory dataSourceFactory = new ClickhouseDataSourceFactory();
        dataSourceFactory.setProperties(ckProps);
        Environment environment = new Environment("clickhouse", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(PlanStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(PlanStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", Types.STRING, Types.POJO(PlanStatOfDayDWD.class)));
        lastReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("lastReduceState", Types.STRING, Types.POJO(PlanStatOfHourDWD.class)));
    }

    @Override
    public void processElement(AdDataOfHourODS element, KeyedProcessFunction<Long, AdDataOfHourODS, PlanStatOfHourDWD>.Context context,
                               Collector<PlanStatOfHourDWD> collector) throws Exception {
        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
        long now = System.currentTimeMillis();
        long adId = element.getAdId();
        LocalDate today = LocalDate.now();

        String lastQueryDay = lastQueryDayState.value();
        // 从 maxCompute拉取指定 广告的历史数据
        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(today))) {
            try (SqlSession session = sqlSessionFactory.openSession()) {
                PlanStatOfDayDWDMapper mapper = session.getMapper(PlanStatOfDayDWDMapper.class);
                Map<String, PlanStatOfDayDWD> historyData = mapper.lastReduceResult(adId, null, DateUtil.formatLocalDate(today.minusDays(1L)), 60)
                        .stream()
                        .peek(value -> {
                            value.setAdIds(PlanUtil.parseAdStr(value.getAdIdsStr()));
                            value.setAdGroupMap(PlanUtil.parseAdGroupMapStr(value.getAdGroupMapStr()));
                        })
                        .collect(Collectors.toMap(PlanStatOfDayDWD::getStatDay, data -> data, (val1, val2) -> val2));
                historyReduceState.clear();
                historyReduceState.putAll(historyData);
                lastQueryDayState.update(DateUtil.formatLocalDate(today));
            }
        }
        PlanStatOfDayDWD yesterdayReduceData = null;
        for (int i = 1; i < 60; i++) {
            LocalDate day = statDay.minusDays(i);
            yesterdayReduceData = historyReduceState.get(DateUtil.formatLocalDate(day));
            if (yesterdayReduceData != null) {
                break;
            }
        }

        PlanStatOfHourDWD lastReduceData = lastReduceState.get(element.getStatDay());

        PlanStatOfHourDWD newStatData = PlanStatOfHourDWD.reduce(yesterdayReduceData, lastReduceData, element, now);
        lastReduceState.put(newStatData.getStatDay(), newStatData);
        collector.collect(newStatData);

        lastReduceState.remove(DateUtil.formatLocalDate(statDay.minusDays(15L)));
    }
}
